"""This is the demo script to show how to manage attributes, attribute forms
and expressions. This script will not work without replacing parameters
with real values. Its basic goal is to present what can be done with
this module and to ease its usage.
"""
from mstrio.connection import Connection, get_connection
from mstrio.modeling.schema import (
    Attribute,
    AttributeDisplays,
    AttributeForm,
    AttributeSort,
    AttributeSorts,
    DataType,
    FormReference,
    list_attributes,
    ObjectSubType,
    Relationship,
    SchemaManagement,
    SchemaObjectReference,
    SchemaUpdateType
)
from mstrio.modeling.expression import (
    Expression,
    ExpressionFormat,
    Token,
    ColumnReference,
    Constant,
    Operator,
    FactExpression,
    Variant,
    VariantType,
    Function
)
from workflows.get_all_columns_in_table import list_table_columns

# Following variables are defining basic attributes
PROJECT_NAME = '<Project_name>'  # Insert name of project here
ATTRIBUTE_NAME = '<Attribute_name>'  # Insert name of edited attribute here
ATTRIBUTE_ID = '<Attribute_ID>'  # Insert ID of edited attribute here
ATTRIBUTE_NEW_NAME = '<Attribute_name>'  # Insert new name of edited attribute here
ATTRIBUTE_DESCRIPTION = '<Attribute_desc>'  # Insert new description of edited attribute here
TABLE_NAME = '<Table_name>'  # Insert table name here
ATTRIBUTE_FORM_NAME = '<Attribute_name>'  # Insert attribute form name here
ATTRIBUTE_FORM_ALTERED_NAME = '<Attribute_name>'  # Insert altered attribute form name here
ATTRIBUTE_FORM_DESCRIPTION = '<Attribute_desc>'  # Insert altered attribute form description here
FOLDER_ID = '<Folder_ID>'  # Insert folder ID here

conn = get_connection(workstationData, PROJECT_NAME)

# Function list_table_columns() might be useful to get table columns information
# needed to create/alter an Attribute
table_columns = list_table_columns(conn)

# Example attribute data.
# Parts of this dictionary will be used in the later parts of this demo script
ATTRIBUTE_DATA = {
    'name': 'Demo Attribute',
    'sub_type': ObjectSubType.ATTRIBUTE,
    'destination_folder': FOLDER_ID,
    'forms': [
        AttributeForm.local_create(
            conn,
            name='ID',
            alias='day_date',
            category='ID',
            display_format=AttributeForm.DisplayFormat.DATE,
            data_type=DataType(
                type=DataType.Type.TIME_STAMP,
                precision=8,
                scale=-2147483648,
            ),
            expressions=[
                FactExpression(
                    expression=Expression(
                        tree=ColumnReference(
                            column_name='day_date',
                            object_id='<Object_ID>',
                        ),
                    ),
                    tables=[
                        SchemaObjectReference(
                            name='DAY_CTR_SLS',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<Object_ID>',
                        ),
                        SchemaObjectReference(
                            name='LU_DAY',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<Object_ID>',
                        ),
                    ],
                ),
            ],
            lookup_table=SchemaObjectReference(
                name='LU_DAY',
                sub_type=ObjectSubType.LOGICAL_TABLE,
                object_id='<Object_ID>',
            ),
        ),
        AttributeForm.local_create(
            conn,
            name='COST',
            alias='total_cost',
            category='COST',
            display_format=AttributeForm.DisplayFormat.NUMBER,
            data_type=DataType(
                type=DataType.Type.FLOAT,
                precision=8,
                scale=-2147483648,
            ),
            expressions=[
                FactExpression(
                    expression=Expression(
                        tree=ColumnReference(
                            column_name='tot_cost',
                            object_id='<Object_ID>',
                        ),
                    ),
                    tables=[
                        SchemaObjectReference(
                            name='DAY_CTR_SLS',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<Object_ID>',
                        )
                    ],
                ),
            ],
            lookup_table=SchemaObjectReference(
                name='DAY_CTR_SLS',
                sub_type=ObjectSubType.LOGICAL_TABLE,
                object_id='<Object_ID>',
            ),
        ),
    ],
    'key_form': FormReference(name='ID'),
    'displays': AttributeDisplays(
        report_displays=[FormReference(name='ID')],
        browse_displays=[FormReference(name='ID')],
    ),
    'attribute_lookup_table': SchemaObjectReference(
        name='LU_DAY',
        sub_type=ObjectSubType.LOGICAL_TABLE,
        object_id='<Object_ID>',
    ),
}

# Attributes management
# Get list of attributes, with examples of different conditions
# with expressions represented as trees
list_of_all_atrs = list_attributes(connection=conn)
list_of_limited_atrs = list_attributes(connection=conn, limit=10)
list_of_atrs_by_name = list_attributes(connection=conn, name=ATTRIBUTE_NAME)
list_of_atrs_by_id = list_attributes(connection=conn, id=ATTRIBUTE_ID)
list_of_atrs_as_trees = list_attributes(conn, show_expression_as=ExpressionFormat.TREE)

# list attributes with expressions represented as tokens
list_of_atrs_as_tokens = list_attributes(conn, show_expression_as=ExpressionFormat.TOKENS)

# Get specific attribute by id or name with expression represented as trees
attr = Attribute(connection=conn, id=ATTRIBUTE_ID)
attr = Attribute(connection=conn, name=ATTRIBUTE_NAME)
attr = Attribute(conn, name=ATTRIBUTE_NAME, show_expression_as=ExpressionFormat.TREE)

# Get an attribute with expression represented as tokens
attr = Attribute(conn, id=ATTRIBUTE_ID, show_expression_as=ExpressionFormat.TOKENS)

# Listing properties
properies = attr.list_properties()

# Create attribute and get it with expression represented as tree
attr = Attribute.create(connection=conn, **ATTRIBUTE_DATA)
attr = Attribute.create(
    connection=conn,
    name=ATTRIBUTE_DATA['name'],
    sub_type=ATTRIBUTE_DATA['sub_type'],
    destination_folder=ATTRIBUTE_DATA['destination_folder'],
    forms=ATTRIBUTE_DATA['forms'],
    key_form=ATTRIBUTE_DATA['key_form'],
    displays=ATTRIBUTE_DATA['displays'],
    attribute_lookup_table=ATTRIBUTE_DATA['attribute_lookup_table'],
)
attr = Attribute.create(
    connection=conn, **ATTRIBUTE_DATA, show_expression_as=ExpressionFormat.TREE
)

# Create an attribute and get it with expression represented as tokens
attr = Attribute.create(conn, **ATTRIBUTE_DATA, show_expression_as=ExpressionFormat.TOKENS)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Alter attributes
attr.alter(name=ATTRIBUTE_NEW_NAME, description=ATTRIBUTE_DESCRIPTION)

# Get displays and sorts
displays = attr.displays
sorts = attr.sorts

# Alter displays and sorts
attr.alter(
    displays=AttributeDisplays(
        report_displays=[FormReference(name='ID'), FormReference(name='COST')],
        browse_displays=[FormReference(name='ID')],
    )
)
attr.alter(
    sorts=AttributeSorts(
        report_sorts=[
            AttributeSort(FormReference(name='ID'), ascending=True),
            AttributeSort(FormReference(name='COST'))
        ],
        browse_sorts=[AttributeSort(FormReference(name='COST'))],
    )
)

# *Relationship management
# Listing relationship candidates
attr_item = Attribute(conn, name=ATTRIBUTE_NAME)
candidates = attr_item.list_relationship_candidates(already_used=False, to_dictionary=True)

# Listing tables
tables = attr_item.list_tables()

# Select table with given TABLE_NAME from tables
[choosen_table] = [tab for tab in tables if tab.name == TABLE_NAME]

# Get candidates from choosen table
candidate_attrs = candidates[choosen_table.name]

# Add child and parent
relationship_child: SchemaObjectReference = candidate_attrs[0]
relationship_child_1: SchemaObjectReference = candidate_attrs[1]
relationship_parent: SchemaObjectReference = candidate_attrs[2]
relationship_child_table: SchemaObjectReference = choosen_table

attr_item.add_child(
    relationship_child,
    relationship_type=Relationship.RelationshipType.ONE_TO_MANY,
    table=relationship_child_table
)
attr_item.add_child(
    joint_child=[relationship_child, relationship_child_1],
    relationship_type=Relationship.RelationshipType.MANY_TO_MANY,
    table=relationship_child_table
)
attr_item.add_parent(relationship_parent, table=choosen_table)

# Remove child and parent
attr_item.remove_child(relationship_child)
attr_item.remove_child(joint_child=[relationship_child, relationship_child_1])
attr_item.remove_parent(relationship_parent)

# *Attribute forms management
# Add a form to the attribute
attribute_form = AttributeForm.local_create(
    conn,
    name='SALES',
    alias='tot_unit_sales',
    category='SALES',
    display_format=AttributeForm.DisplayFormat.NUMBER,
    data_type=DataType(
        type=DataType.Type.FLOAT,
        precision=8,
        scale=-2147483648,
    ),
    expressions=[
        FactExpression(
            expression=Expression(
                tree=ColumnReference(
                    column_name='tot_unit_sales',
                    object_id='<Object_ID>',
                ),
            ),
            tables=[
                SchemaObjectReference(
                    name='DAY_CTR_SLS',
                    sub_type=ObjectSubType.LOGICAL_TABLE,
                    object_id='<Object_ID>',
                )
            ],
        ),
    ],
    lookup_table=SchemaObjectReference(
        name='DAY_CTR_SLS',
        sub_type=ObjectSubType.LOGICAL_TABLE,
        object_id='<Object_ID>',
    ),
)

attr.add_form(form=attribute_form)

# Get form with specified id. This retrieves the form from the local
# Attribute object, not from the server
attr_form = attr.get_form(name=ATTRIBUTE_FORM_NAME)
# Alter form
attr.alter_form(
    form_id=attr_form.id, name=ATTRIBUTE_FORM_ALTERED_NAME, description=ATTRIBUTE_FORM_DESCRIPTION
)
attr.alter_form(form_id=attr_form.id, name=ATTRIBUTE_FORM_ALTERED_NAME)

# Removing form with specified id
attr.remove_form(form_id=attr_form.id)

# *AttributeForm's expression management
# Add a fact expression to the attribute form

# create a fact expression with expression specified as tokens
fact_expression = FactExpression(
    expression=Expression(
        tokens=[
            Token(
                type=Token.Type.COLUMN_REFERENCE,
                value='order_date',
                target=SchemaObjectReference(
                    object_id='<Object_ID>',
                    name='order_date',
                    sub_type=ObjectSubType.COLUMN,
                ),
                level=Token.Level.RESOLVED,
                state=Token.State.INITIAL,
            ),
            Token(
                type=Token.Type.END_OF_TEXT,
                value='',
                level=Token.Level.RESOLVED,
                state=Token.State.INITIAL,
            ),
        ],
    ),
    tables=[
        SchemaObjectReference(
            name='table 1 name',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='table 1 id',
        ),
        SchemaObjectReference(
            name='table 2 name',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='table 2 id',
        ),
    ],
)

# create a fact expression with expression specified as tree
fact_expression = FactExpression(
    expression=Expression(
        tree=ColumnReference(
            column_name='DAY_DATE',
            object_id='<Object_ID>',
        ),
    ),
    tables=[
        SchemaObjectReference(
            name='LU_DAY',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<Object_ID>',
        ),
    ],
)

# Add new fact expression to a form
form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)
attr.add_fact_expression(form_id=form.id, expression=fact_expression)

# Create a new expression locally using tree format
new_expression = Expression(
    tree=Operator(
        function=Function.ADD,
        children=[
            ColumnReference(
                column_name='ly_date_day',
                object_id='<Object_ID>',
            ),
            Constant(variant=Variant(VariantType.INT32, '1')),
        ],
    ),
)

attr_form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)

# Get a fact expression from form by using form name
fact_expression, *_ = [f for f in attr_form.expressions if f.expression.text == 'DAY_DATE']

# Alter the fact expression
attr.alter_fact_expression(
    form_id=attr_form.id, fact_expression_id=fact_expression.id, expression=new_expression
)

# Remove the fact expression from the the attribute form
attr_form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)
fact_expression, *_ = attr_form.expressions
attr.remove_fact_expression(form_id=form.id, fact_expression_id=fact_expression.id)

# Deleting attributes
attr.delete(force=True)



"""This is the demo script to show how to manage tables.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.datasources.datasource_instance import list_connected_datasource_instances
from mstrio.modeling import DataType, SchemaManagement, SchemaUpdateType
from mstrio.modeling.schema.helpers import (
    ObjectSubType, PhysicalTableType, SchemaObjectReference, TableColumn, TableColumnMergeOption
)
from mstrio.modeling.schema.table import (
    list_datasource_warehouse_tables,
    list_logical_tables,
    list_namespaces,
    list_physical_tables,
    list_tables_prefixes,
    LogicalTable,
    PhysicalTable,
)

PROJECT_ID = "<project_id>"  # Project ID to connect to
PROJECT_NAME = "<project_name>"  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get a list of table prefixes
table_prefixes = list_tables_prefixes(connection=conn)
print(table_prefixes)

# Define a variable, which can be later used in a script
PHYSICAL_TABLE_PREFIX = "<physical_table_prefix>"  # Insert prefix of the Table here

# Get a list of physical tables
physical_tables = list_physical_tables(connection=conn, project_id=PROJECT_ID)
print(physical_tables)

# Get a list of logical tables
logical_tables = list_logical_tables(connection=conn)
print(logical_tables)

# Get datasource warehouse table
# 1. List connected datasource instances
connected_datasource_instances = list_connected_datasource_instances(connection=conn)
print(connected_datasource_instances)

# Define a variable, which can be later used in a script
DATASOURCE_ID = "<datasource_id>"
DATASOURCE_NAME = "<datasource_name>"

# 2. Get a datasource by name
datasource = [
    datasource for datasource in list_connected_datasource_instances(connection=conn)
    if datasource.name == DATASOURCE_NAME
][0]
print(datasource)

# 3. List namespaces
namespaces = list_namespaces(connection=conn, id=DATASOURCE_ID)
print(namespaces)

# Define a variable, which can be later used in a script
NAMESPACE_ID = "<namespace_id>"
NAMESPACE_NAME = "<namespace_name>"

# 4. Get a namespace by name for a specific datasource
namespace = [
    namespace for namespace in list_namespaces(connection=conn, id=DATASOURCE_ID)
    if namespace.get("name") == NAMESPACE_NAME
][0]
print(namespace)

# 5. List available Warehouse Tables in a specific datasource within a given namespace
warehouse_tables = list_datasource_warehouse_tables(
    connection=conn, datasource_id=DATASOURCE_ID, namespace_id=NAMESPACE_ID
)
print(warehouse_tables)

# Define a variables, which can be later used in a script
WAREHOUSE_TABLE_NAME = "<warehouse_table_name>"  # Insert name of the Table here

# Get a Warehouse Tables with the name WAREHOUSE_TABLE_NAME
# from the specified DATASOURCE_ID and NAMESPACE_ID
lu_item_table = list_datasource_warehouse_tables(
    connection=conn,
    datasource_id=DATASOURCE_ID,
    namespace_id=NAMESPACE_ID,
    name=WAREHOUSE_TABLE_NAME
)[0]

print(lu_item_table.list_columns())

# List dependent logical tables of a Warehouse Table.
print(lu_item_table.list_dependent_logical_tables())

# Add a Warehouse Table to a project. This will automatically create both a
# logical table object and physical table object on the server

# Define a variable, which can be later used in a script
NEW_LOGICAL_TABLE_NAME = "<new_logical_table_name>"  # Insert name of the logical table

lu_item_table = list_datasource_warehouse_tables(
    connection=conn,
    datasource_id=DATASOURCE_ID,
    namespace_id=NAMESPACE_ID,
    name=WAREHOUSE_TABLE_NAME
)[0]

# List logical tables
logical_tables = list_logical_tables(conn)
print(logical_tables)

# Define variables, which can be later used in a script
LOGICAL_TABLE_ID = "<logical_table_id>"  # Insert ID of the logical table
LOGICAL_TABLE_NAME = "<logical_table_name>"  # Insert name of the logical table

# These tables can then be retrieved from the server on demand.
# Get logical table by id
logical_table = LogicalTable(connection=conn, id=LOGICAL_TABLE_ID)
print(logical_table)

# Get logical table by name
logical_table = LogicalTable(connection=conn, name=LOGICAL_TABLE_NAME)
print(logical_table)

# List columns of logical table
print(logical_table.list_columns())

# List physical tables in a project
list_physical_tables(conn, project_id=PROJECT_ID)

# Define a variable, which can be later used in a script
PHYSICAL_TABLE_ID = "<physical_table_id>"

# Physical table can be retrieved by its id
physical_table = PhysicalTable(connection=conn, id=PHYSICAL_TABLE_ID)
print(physical_table)

# List dependent logical tables of a physical table
print(physical_table.list_dependent_logical_tables())

# Define variable, which can be later used in a script
# The variable below is needed when altering or creating a logical table
LOGICAL_TABLE_DESCRIPTION = (
    "<logical_table_description>"  # Insert description of the logical table
)

# Remove all logical tables added for a specific Warehouse Table from a project.
# All dependent tables will be listed before deletion.
lu_item_table.delete_from_project(force=True)

# List all logical tables based on physical tables of type normal
normal_tables = list_logical_tables(connection=conn, table_type=PhysicalTableType.NORMAL)
print(normal_tables)

# Define a variable, which can be later used in a script
NORMAL_TABLE_NAME = "<normal_table_name>"

# Get a table by the name
normal_table = LogicalTable(conn, name=NORMAL_TABLE_NAME)
print(normal_table)

# List all logical tables based on physical tables of type SQL
sql_tables = list_logical_tables(connection=conn, table_type=PhysicalTableType.SQL)
print(sql_tables)

# Define a variable, which can be later used in a script
SQL_TABLE_NAME = "<sql_table_name>"

# Get a table by the name
sql_table = LogicalTable(conn, name=SQL_TABLE_NAME)
print(sql_table)

# List all logical tables based on physical tables of type warehouse_partition
warehouse_partition_tables = list_logical_tables(
    connection=conn, table_type=PhysicalTableType.WAREHOUSE_PARTITION
)
print(warehouse_partition_tables)

# List properties for the first logical table in a list
print(logical_tables[0].list_properties())

# List properties for the third logical table in a list
print(logical_tables[2].list_properties())

# Remove logical table
logical_tables[0].delete(force=True)

# Alter table

# Define variables, which can be later used in a script
# The variable below is needed when altering a logical table
LOGICAL_TABLE_NAME_ALTERED = "<logical_table_name>"  # Insert altered Table name
# The variable below is needed when altering or creating a logical table
LOGICAL_TABLE_DESCRIPTION = (
    "<logical_table_description>"  # Insert description of the logical table
)
# Physical Table
PHYSICAL_TABLE_NAME_ALTERED = "<physical_table_name>"  # Insert altered Table name

# Alter logical table based on physical table of type normal
normal_table.alter(name=LOGICAL_TABLE_NAME_ALTERED)  # logical table name
normal_table.alter(is_true_key=True)
normal_table.alter(logical_size=10)
normal_table.alter(is_logical_size_locked=True)
normal_table.alter(physical_table_object_name=PHYSICAL_TABLE_NAME_ALTERED)
normal_table.alter(physical_table_prefix=PHYSICAL_TABLE_PREFIX)

# Alter logical table based on physical table of type SQL
sql_table.alter(name=LOGICAL_TABLE_NAME)  # logical table name
sql_table.alter(is_true_key=True)
sql_table.alter(logical_size=10)
sql_table.alter(is_logical_size_locked=True)
sql_table.alter(enclose_sql_in_parentheses=True)
sql_table.alter(physical_table_object_name=PHYSICAL_TABLE_NAME_ALTERED)

# You cannot alter logical tables based on physical tables of type warehouse_partition!

# Update structure of physical table of type normal
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_ANY)
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE)
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE)

# Update structure of physical table of type SQL
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_ANY)
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE)
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE)

# Delete a physical table.
physical_table.delete(force=True)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=PROJECT_ID)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])



# flake8: noqa
from .object import list_objects, Object
from .predefined_folders import PredefinedFolders

# isort: off
from .folder import (
    Folder, get_my_personal_objects_contents, get_predefined_folder_contents, list_folders
)
# isort: on
from .search_enums import CertifiedStatus, SearchDomain, SearchPattern, SearchResultsFormat
from .search_operations import (
    CertifiedStatus,
    full_search,
    get_search_results,
    get_search_suggestions,
    quick_search,
    quick_search_from_object,
    SearchDomain,
    SearchObject,
    SearchPattern,
    SearchResultsFormat,
    start_full_search
)
from .shortcut import get_shortcuts, Shortcut, ShortcutInfoFlags



from collections import defaultdict
from enum import auto
import logging
from typing import List, Optional, TYPE_CHECKING, Union

from mstrio import config
from mstrio.api import objects, transmitters
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable, fetch_objects, get_args_from_func, get_default_args_from_func
)
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class RecipientFieldType(AutoName):
    TO = auto()
    CC = auto()
    BCC = auto()


class TransmitterDeliveryType(AutoName):
    EMAIL = auto()
    FTP = auto()
    FILE = auto()
    PRINT = auto()
    IPHONE = auto()
    IPAD = auto()
    ANDROID = auto()
    UNSUPPORTED = auto()


class EmailTransmitterProperties(Dictable):
    """Representation of email transmitter properties object. Those properties
    are specific for Email Transmitters.

    Attributes:
        sender_display_name: name used in the email headers to indicate message
            sender
        sender_email_address: address used in the email headers to indicate
            message sender
        reply_to_display_name: name used in the email headers to indicate
            destination for replies
        reply_to_email_address: address used in the email headers to indicate
            destination for replies
        recipient_field_type: default field in which recipient address appear
            (eg. to/cc/bcc). Default value is `RecipientFieldType.TO`.
        save_message_to_file: specifies whether to save message output to file,
            default value is `False`
        send_message_via_smtp: specifies whether to send messages to recipients
            via SMTP, default value is `False`
        save_file_path: the folder path to save the email output,
            default is `None`
        notify_on_success: specifies whether to request notification on success,
            default is `False`
        notify_on_failure: specifies whether to request notification on failure,
            default is `False`
        notification_email_address: email address for notification, default is
            `None`
    """
    _DELETE_NONE_VALUES_RECURSION = False
    _FROM_DICT_MAP = {'recipient_field_type': RecipientFieldType}

    def __init__(
        self,
        sender_display_name: str,
        sender_email_address: str,
        reply_to_display_name: str,
        reply_to_email_address: str,
        recipient_field_type: RecipientFieldType = RecipientFieldType.TO,
        save_message_to_file: bool = False,
        send_message_to_file: bool = False,
        send_message_via_smtp: bool = False,
        save_file_path: Optional[str] = None,
        notify_on_success: bool = False,
        notify_on_failure: bool = False,
        notification_email_address: Optional[str] = None
    ):
        self.sender_display_name = sender_display_name
        self.sender_email_address = sender_email_address
        self.reply_to_display_name = reply_to_display_name
        self.reply_to_email_address = reply_to_email_address
        self.recipient_field_type = recipient_field_type
        self.save_message_to_file = save_message_to_file
        self.send_message_to_file = send_message_to_file
        self.send_message_via_smtp = send_message_via_smtp
        self.save_file_path = save_file_path
        self.notify_on_success = notify_on_success
        self.notify_on_failure = notify_on_failure
        self.notification_email_address = notification_email_address


@method_version_handler('11.3.0100')
def list_transmitters(
    connection: "Connection", to_dictionary: bool = False, limit: Optional[int] = None, **filters
) -> Union[List["Transmitter"], List[dict]]:
    """Get all transmitters as list of Transmitter objects or
    dictionaries.

    Optionally filter the transmitters by specifying filters.

    Args:
        connection(object): MicroStrategy connection object
        to_dictionary: If True returns a list of transmitter dicts,
            otherwise returns a list of transmitter objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'delivery_type']
    """
    return Transmitter._list_transmitters(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@class_version_handler('11.3.0100')
class Transmitter(Entity, DeleteMixin):
    """Object representation of MicroStrategy Transmitter object

    Attributes:
        name: transmitter's name
        id: transmitter's id
        description: transmitter's description
        delivery_type: type of the transmitter
        email_transmitter_properties: properties specific to email transmitters
            (available only when `delivery_type` equals
            `TransmitterDeliveryType.EMAIL`)
    """
    _DELETE_NONE_VALUES_RECURSION = False
    _OBJECT_TYPE = ObjectTypes.SUBSCRIPTION_TRANSMITTER
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'delivery_type': TransmitterDeliveryType,
        'email_transmitter_properties': EmailTransmitterProperties.from_dict
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl'
        ): objects.get_object_info,
        ('id', 'name', 'description', 'delivery_type',
         'email_transmitter_properties'): transmitters.get_transmitter
    }
    _API_DELETE = staticmethod(transmitters.delete_transmitter)
    _API_PATCH: dict = {
        ("name", "description",
         "email_transmitter_properties"): (transmitters.update_transmitter, "put")
    }
    _PATCH_PATH_TYPES = {"name": str, "description": str, "email_transmitter_properties": dict}

    def __init__(
        self, connection: "Connection", id: Optional[str] = None, name: Optional[str] = None
    ):
        """Initialize Transmitter object.

        Note:
            In case both `id` and `name` are provided, then `id` is used when
            fetching.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing transmitter
            name (str, optional): Name of the pre-existing transmitter

        Raises:
            `ValueError` when neither `id` nor `name` are specified.
        """

        if id is None and name is None:
            raise ValueError("Please specify either 'id' or 'name' parameter in the constructor.")

        if id is None:
            objects_info = Transmitter._list_transmitters(
                connection=connection,
                name=name,
                to_dictionary=True,
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Transmitter: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._delivery_type = TransmitterDeliveryType(kwargs["delivery_type"]
                                                      ) if kwargs.get("delivery_type") else None
        self.email_transmitter_properties = EmailTransmitterProperties.from_dict(
            kwargs["email_transmitter_properties"]
        ) if kwargs.get("email_transmitter_properties") else None

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        delivery_type: Union[str, TransmitterDeliveryType],
        description: Optional[str] = None,
        email_transmitter_properties: Optional[Union[dict, EmailTransmitterProperties]] = None
    ) -> "Transmitter":
        """Create transmitter.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            delivery_type: type of the transmitter
            name (str): transmitter's name
            description (str): transmitter's description
            email_transmitter_properties (dict or object): properties specific
                to email transmitter. In transmitter with type `email` those
                properties must be provided. Otherwise they cannot be provided.

        Returns:
            `Transmitter` object
        """
        delivery_type = get_enum_val(delivery_type, TransmitterDeliveryType)
        email_transmitter_properties = email_transmitter_properties.to_dict() if isinstance(
            email_transmitter_properties, EmailTransmitterProperties
        ) else email_transmitter_properties
        body = {
            'name': name,
            'description': description,
            'deliveryType': delivery_type,
            'emailTransmitterProperties': email_transmitter_properties
        }

        res = transmitters.create_transmitter(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created transmitter named: '{res.get('name')}' "
                f"with ID: '{res.get('id')}'"
            )
        return cls.from_dict(res, connection)

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        email_transmitter_properties: Optional[Union[dict, EmailTransmitterProperties]] = None
    ):
        """Alter transmitter properties.

        Args:
            name (str): transmitter's name
            description (str): transmitter's description
            email_transmitter_properties (dict or object): properties specific
                to email transmitter. Only in transmitter with type `email`
                altering those properties is possible
        """
        email_transmitter_properties = email_transmitter_properties.to_dict() if isinstance(
            email_transmitter_properties, EmailTransmitterProperties
        ) else email_transmitter_properties
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = defaultdict(dict)
        for property_key in defaults_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_transmitters(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: Optional[int] = None,
        **filters
    ) -> Union[List["Transmitter"], List[dict]]:
        """Helper method for listing transmitters."""
        objects = fetch_objects(
            connection=connection,
            api=transmitters.get_transmitters,
            limit=limit,
            filters=filters,
            dict_unpack_value="transmitters",
        )

        if to_dictionary:
            return objects
        return [Transmitter.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def delivery_type(self):
        return self._delivery_type



from enum import auto
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from mstrio.access_and_security.security_filter import ObjectReference
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import Dictable

if TYPE_CHECKING:
    from mstrio.connection import Connection


class ParameterType(AutoName):
    CONSTANT = auto()
    OBJECT_REFERENCE = auto()
    EXPRESSION = auto()
    PROMPT = auto()
    ARRAY = auto()
    DYNAMIC_DATE_TIME = auto()


class ParameterBase(Dictable):

    TYPE = None
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self):
        self.parameter_type = self.TYPE


class ConstantType(AutoName):
    INT_32 = "int32"
    INT_64 = "int64"
    DOUBLE = auto()
    DATE = auto()
    TIME = auto()
    STRING = auto()


class ConstantParameter(ParameterBase):

    TYPE = ParameterType.CONSTANT
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, type: Union["ConstantType", str], value: str):
        """Constant parameter for predicate.

        Args:
            type (enum): type of constant. Possible values are defined in enum
                `ConstantType`.
            value (str): value of constant.
        """
        super().__init__()
        self.constant = {"type": get_enum_val(type, ConstantType), "value": value}

    @classmethod
    def from_dict(cls, source: Dict[str, Any], connection: Optional["Connection"] = None):
        source["type"] = source.get('constant', {}).get('type')
        source["value"] = source.get('constant', {}).get('value')
        return super().from_dict(source, connection)


class ObjectReferenceParameter(ParameterBase):

    TYPE = ParameterType.OBJECT_REFERENCE
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, target: "ObjectReference"):
        super().__init__()
        self.target = target

    _FROM_DICT_MAP = {"target": ObjectReference.from_dict}


class ExpressionParameter(ParameterBase):

    TYPE = ParameterType.OBJECT_REFERENCE
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, expression: dict):
        super().__init__()
        self.expression = expression


class PromptParameter(ParameterBase):

    TYPE = ParameterType.PROMPT
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, prompt: "ObjectReference"):
        super().__init__()
        self.prompt = prompt

    _FROM_DICT_MAP = {"prompt": ObjectReference.from_dict}


class DynamicDateTimeParameter(ParameterBase):

    TYPE = ParameterType.DYNAMIC_DATE_TIME
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, dynamic_date_time: dict):
        super().__init__()
        # TODO think about some object for dynamic_date_structure
        self.dynamic_date_time = dynamic_date_time


class ConstantArrayParameter(ParameterBase):

    TYPE = ParameterType.ARRAY
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(self, type: Union["ConstantType", str], values: List[str]):
        super().__init__()
        self.constants_type = type if isinstance(type, ConstantType) else ConstantType(type)
        self.constants = values

    @classmethod
    def from_dict(
        cls,
        source: Dict[str, Any],
        connection: Optional["Connection"] = None,
        to_snake_case: bool = True
    ):
        source["type"] = source.get('constantsType' if to_snake_case else 'constants_type')
        source["values"] = source.get('constants')
        return super().from_dict(source, connection)
